package core;

import utils.exceptions.FileParseException;
import java.io.IOException;
import java.util.StringTokenizer;
import utils.io.BufferedReaderExtended;
import utils.UtilsIO;
import utils.UtilsLogging;
import utils.exceptions.SyntaxException;

/**
 *
 * @author Jason Koukas
 */
public class SRTBlock extends SubtitleBlock {

    public static final int FIRST_SUBTITLE_INDEX_MIN = 0;
    public static final int FIRST_SUBTITLE_INDEX_MAX = 4999;
    public static final int FIRST_SUBTITLE_INDEX_DEFAULT = 1;
    public static final int SUBTITLE_INDEX_SPECIAL = 9999;
    public static final int HOURS_MAX_VALUE = 595;
    private static final int[] MILLIS_PER_TIME_UNITS = new int[]{TimeConstants.MILLIS_PER_HOUR, TimeConstants.MILLIS_PER_MINUTE, TimeConstants.MILLIS_PER_SECOND};
    //
    private int m_index;
    private String m_metaData; //optional meta-data

    public SRTBlock() {
        //m_metaData = "";
    }

    public void setIndex(int index) {
        m_index = index;
    }

    public int getIndex() {
        return m_index;
    }

    @Override
    protected void clearDataSubclass() {
        m_index = 0;
        m_metaData = null;
    }

    @Override
    protected char getLinesDelimiter() {
        return '\n';
    }

    @Override
    public SubtitleType getType() {
        return SubtitleType.SRT_TYPE;
    }

    @Override
    public String getTimestampString() {
        return millisecondsToTime(m_timeStart);
    }

    @Override
    public String toStringRepresentation() {
        String ret = m_index + "\n" + millisecondsToTime(m_timeStart) + " --> " + millisecondsToTime(m_timeEnd);
        if (m_metaData != null) {
            ret += "  " + m_metaData;
        }
        ret += "\n" + toString() + "\n";
        return ret;
    }

    public boolean readBlock(BufferedReaderExtended br)
            throws IOException, FileParseException {
        clearData(); //for safety because one might assume that the block's clear

        try {
            String line = UtilsIO.getNextLine(br, true);
            if (line == null) {
                return false;
            }
            m_index = verifyInteger(line);

            line = demandLine(br, true);
            verifyAssignTimestampsLine(line);

            line = demandLine(br, true);
            verifyNotTimestamps(line);
            m_subtitleLines.clear();
            m_subtitleLines.add(line); //we assume that all have one line at least

            boolean gapFound = false;
            while ((line = UtilsIO.getNextLine(br, false)) != null) {
                if ("".equals(line)) {
                    gapFound = true;
                    continue;
                }
                try {
                    verifyInteger(line);
                    if (gapFound == true) {
                        br.unreadLastLine();
                        break;
                    }
                } catch (SyntaxException sexc) {
                    gapFound = false;
                }
                verifyNotTimestamps(line);
                m_subtitleLines.add(line);
            }
            return true;
        } catch (IOException ex) {
            clearData();
            throw ex;
        } catch (SyntaxException ex) {
            UtilsLogging.log(getClass(), ex);
            clearData();
            throw new FileParseException(ex.getMessage(), br.getLinesRead(), br.getLastReadLine());
        }
    }

    @Override
    public SubtitleBlock clone() {
        SRTBlock clone = new SRTBlock();
        cloneTo(clone);
        clone.m_index = m_index;
        clone.m_metaData = m_metaData;
        return clone;
    }

    public static boolean validateTimeString(String value) {
        StringTokenizer st = new StringTokenizer(value, ":");

        if (!validateTokenTime(st, HOURS_MAX_VALUE + 1) || !validateTokenTime(st, TimeConstants.MINUTES_SECONDS_PER_HOUR_MINUTE) || !st.hasMoreTokens()) {
            return false;
        }
        value = st.nextToken();
        if (st.hasMoreTokens()) {
            return false;
        }
        st = new StringTokenizer(value, ",");

        if (!validateTokenTime(st, TimeConstants.MINUTES_SECONDS_PER_HOUR_MINUTE) || !validateTokenTime(st, TimeConstants.MILLIS_PER_SECOND) || st.hasMoreTokens()) {
            return false;
        }
        return true;
    }

    public static int timeToMilliseconds(String time) throws SyntaxException {
        if (!validateTimeString(time)) {
            throw new SyntaxException(FORMAT_ERR);
        }
        StringTokenizer st = new StringTokenizer(time, ":, ", false); //space is added, to avoid trimming guess

        return timeValuesToMilliseconds(Integer.parseInt(st.nextToken()), Integer.parseInt(st.nextToken()),
                Integer.parseInt(st.nextToken()), Integer.parseInt(st.nextToken()));
    }

    public static int[] millisecondsToSeperateTimeValues(int millis) {
        int[] timeValues = new int[4];
        int lenLimit = timeValues.length - 1;
        int millisPerUnit;
        for (int i = 0; i < lenLimit; i++) {
            millisPerUnit = MILLIS_PER_TIME_UNITS[i];
            timeValues[i] = millis / millisPerUnit;
            millis = millis % millisPerUnit;
        }
        timeValues[lenLimit] = millis;
        return timeValues;
    }

    public static String millisecondsToTime(int millis) {
//        if (millis < 0) { //Shouldn't happen..
//            throw new FormattingException(FORMAT_ERR);
//        }
        int[] vals = millisecondsToSeperateTimeValues(millis);

        return padUnitValue(vals[0], 2) + ":" + padUnitValue(vals[1], 2) + ":" + padUnitValue(vals[2], 2) + "," + padUnitValue(vals[3], 3);
    }

    public static int timeValuesToMilliseconds(int hours, int minutes, int seconds, int millis) {
        return hours * TimeConstants.MILLIS_PER_HOUR
                + minutes * TimeConstants.MILLIS_PER_MINUTE
                + seconds * TimeConstants.MILLIS_PER_SECOND
                + millis;
    }

    //Private Helpful Methods
    private void verifyAssignTimestampsLine(String line) throws SyntaxException {
        int timeStart, timeEnd;
        String metaData;
        StringTokenizer st = new StringTokenizer(line, "->", true);
        timeStart = timeToMilliseconds(demandToken(st));
        verifyNextTokenIs(st, "-");
        verifyNextTokenIs(st, "-");
        verifyNextTokenIs(st, ">");
        line = getRemainingTokensAsLine(st);
        st = new StringTokenizer(line);
        timeEnd = timeToMilliseconds(demandToken(st));
        if (st.hasMoreTokens()) {
            metaData = st.nextToken();
            while (st.hasMoreTokens()) {
                metaData += " " + st.nextToken();
            }
        } else {
            metaData = null; //Unnecessary
        }

        m_timeStart = timeStart;
        m_timeEnd = timeEnd;
        m_metaData = metaData;
    }

    private void verifyNotTimestamps(String line) throws SyntaxException {
        try {
            verifyAssignTimestampsLine(line);
        } catch (SyntaxException se) {
            return;//returns without throwing exception (hopefully..)
        }
        throw new SyntaxException(FORMAT_ERR);
    }

    //Private Static Helpful Methods
    private static String padUnitValue(int timeUnit, int representationMinDigits) {
        String paddingString = "";
        for (int powDecimal = 10; representationMinDigits > 1; --representationMinDigits) {
            if (timeUnit < powDecimal) {
                for (; representationMinDigits > 1; --representationMinDigits) {
                    paddingString += "0";
                }
                break;
            }
            powDecimal *= 10;
        }
        return paddingString + timeUnit;
    }

    private static boolean validateTokenTime(StringTokenizer tokenizer, int maxValueExclusive) {
        try {
            if (!tokenizer.hasMoreTokens()) {
                return false;
            }
            int val = parseIntWithTrim(tokenizer.nextToken());
            if (val < 0 || val >= maxValueExclusive) {
                return false;
            }
            return true;
        } catch (NumberFormatException e) {
            return false;
        }
    }
}
